  /*
  * policy.c
  * Copyright (C) Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
  * 
  * simple-netaid is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  * 
  * simple-netaid is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  * See the GNU General Public License for more details.
  * 
  * You should have received a copy of the GNU General Public License along
  * with this program.  If not, see <http://www.gnu.org/licenses/>.
  * 
  * See the COPYING file.
  */

#include <stdint.h>
#include <syslog.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/wait.h>  // waitpid
#include <libgen.h>     // dirname() and basename()
#include <ctype.h>      // isdigit()	  
#include <errno.h>

#include <simple-netaid/sbuf.h>
#include <simple-netaid/ipaddr.h>
#include <simple-netaid/iproute.h>
#include <simple-netaid/interfaces.h>
#include <simple-netaid/helpers.h>
#include <simple-netaid/wireless.h>

#include "def.h"
#include "config.h"
#include "policy.h"
#include "util.h"
#include "periodic_reminder.h"
#include "iwlist.h"

#include <json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/json_script.h>

#define WIFI_DIR "/etc/network/wifi"
#define PREFIX "/tmp/simple-netaid_"

#define strdup_or_null( str )  (str) != NULL ? strdup(str) : NULL

static struct ubus_request_data req_data;
static struct blob_buf bb_t;

/* Enum for EGI policy order */
enum {
	INTERFACE_DOWN_ID,
	INTERFACE_DOWN_IFNAME,
	__INTERFACE_DOWN_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy interface_down_policy[] =
{
	[INTERFACE_DOWN_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[INTERFACE_DOWN_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
};

/* Enum for EGI policy order */
enum {
	INTERFACE_UP_ID,
	INTERFACE_UP_IFNAME,
	__INTERFACE_UP_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy interface_up_policy[] =
{
	[INTERFACE_UP_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[INTERFACE_UP_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
};

/* Enum for EGI policy order */
enum {
	IFDOWN_ID,
	IFDOWN_IFNAME,
	IFDOWN_TTY,
	__IFDOWN_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy ifdown_policy[] =
{
	[IFDOWN_ID] = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[IFDOWN_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
	[IFDOWN_TTY] = { .name="tty", .type=BLOBMSG_TYPE_INT32 },
};

/* Enum for EGI policy order */
enum {
	IFUP_ID,
	IFUP_IFNAME,
	IFUP_TTY,
	__IFUP_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy ifup_policy[] =
{
	[IFUP_ID] = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[IFUP_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
	[IFUP_TTY] = { .name="tty", .type=BLOBMSG_TYPE_INT32 },
};

/* Enum for EGI policy order */
enum {
	IPADDR_FLUSH_ID,
	IPADDR_FLUSH_IFNAME,
	__IPADDR_FLUSH_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy ipaddr_flush_policy[] =
{
	[IPADDR_FLUSH_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[IPADDR_FLUSH_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
};

/* Enum for EGI policy order */
enum {
	DISCONNECT_ID,
	DISCONNECT_TTY,
	__DISCONNECT_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy disconnect_policy[] =
{
	[DISCONNECT_ID] = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[DISCONNECT_TTY] = { .name="tty", .type=BLOBMSG_TYPE_INT32 },
};

/* Enum for EGI policy order */
enum {
	WIRED_CONNECTION_ID,
	WIRED_CONNECTION_IFNAME,
	WIRED_CONNECTION_TTY,
	__WIRED_CONNECTION_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy wired_connection_policy[] =
{
	[WIRED_CONNECTION_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[WIRED_CONNECTION_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
	[WIRED_CONNECTION_TTY] = { .name="tty", .type=BLOBMSG_TYPE_INT32 },
};
 
/* Enum for EGI policy order */
enum {
	WPA_PASSPHRASE_ID,
	WPA_PASSPHRASE_ESSID,
	WPA_PASSPHRASE_PASSWD,
	WPA_PASSPHRASE_FILENAME,
	__WPA_PASSPHRASE_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy wpa_passphrase_policy[] =
{
	[WPA_PASSPHRASE_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[WPA_PASSPHRASE_ESSID] = { .name="essid", .type=BLOBMSG_TYPE_STRING },
	[WPA_PASSPHRASE_PASSWD] = { .name="passwd", .type=BLOBMSG_TYPE_STRING },
	[WPA_PASSPHRASE_FILENAME] = { .name="filename", .type=BLOBMSG_TYPE_STRING },
};
 
/* Enum for EGI policy order */
enum {
	WPA_SUPPLICANT_ID,
	WPA_SUPPLICANT_IFNAME,
	WPA_SUPPLICANT_FILENAME,
	WPA_SUPPLICANT_TTY,
	__WPA_SUPPLICANT_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy wpa_supplicant_policy[] =
{
	[WPA_SUPPLICANT_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[WPA_SUPPLICANT_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
	[WPA_SUPPLICANT_FILENAME] = { .name="filename", .type=BLOBMSG_TYPE_STRING },
	[WPA_SUPPLICANT_TTY]  = { .name="tty", .type=BLOBMSG_TYPE_INT32 },
};

/* Enum for EGI policy order */
enum {
	WIRELESS_CONNECTION_ID,
	WIRELESS_CONNECTION_IFNAME,
	WIRELESS_CONNECTION_ESSID,
	WIRELESS_CONNECTION_PASSWD,
	WIRELESS_CONNECTION_FILENAME,
	WIRELESS_CONNECTION_TTY,
	__WIRELESS_CONNECTION_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy wireless_connection_policy[] =
{
	[WIRELESS_CONNECTION_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[WIRELESS_CONNECTION_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
	[WIRELESS_CONNECTION_ESSID] = { .name="essid", .type=BLOBMSG_TYPE_STRING },
	[WIRELESS_CONNECTION_PASSWD] = { .name="passwd", .type=BLOBMSG_TYPE_STRING },
	[WIRELESS_CONNECTION_FILENAME] = { .name="filename", .type=BLOBMSG_TYPE_STRING },
	[WIRELESS_CONNECTION_TTY]  = { .name="tty", .type=BLOBMSG_TYPE_INT32 },
};

/* Enum for EGI policy order */
enum {
	UNINSTALL_ID,
	UNINSTALL_FILENAME,
	__UNINSTALL_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy uninstall_policy[] =
{
	[UNINSTALL_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[UNINSTALL_FILENAME] = { .name="filename", .type=BLOBMSG_TYPE_STRING },
};

/* Enum for EGI policy order */
enum {
	SCAN_ACTIVE_WIFIS_ID,
	SCAN_ACTIVE_WIFIS_IFNAME,
	SCAN_ACTIVE_WIFIS_FORMAT,
	__SCAN_ACTIVE_WIFIS_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy scan_active_wifis_policy[] =
{
	[SCAN_ACTIVE_WIFIS_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[SCAN_ACTIVE_WIFIS_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
	[SCAN_ACTIVE_WIFIS_FORMAT]  = { .name="format", .type=BLOBMSG_TYPE_INT32 },
};

/* Enum for EGI policy order */
enum {
	NL_MONITOR_TOUCH_ID,
	NL_MONITOR_TOUCH_UID,
	NL_MONITOR_TOUCH_GID,
	NL_MONITOR_TOUCH_FILENAME,
	__NL_MONITOR_TOUCH_MAX,
};

/* Ubus Policy */
const struct blobmsg_policy nl_monitor_touch_policy[] =
{
	[NL_MONITOR_TOUCH_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[NL_MONITOR_TOUCH_UID]  = { .name="uid", .type=BLOBMSG_TYPE_INT32},
	[NL_MONITOR_TOUCH_GID]  = { .name="gid", .type=BLOBMSG_TYPE_INT32},
	[NL_MONITOR_TOUCH_FILENAME] = { .name="filename", .type=BLOBMSG_TYPE_STRING },
};
 
/* Enum for EGI policy order */
enum {
	SV_DHCPCD_ID,
	SV_DHCPCD_ACTION,
	__SV_DHCPCD_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy sv_dhcpcd_policy[] =
{
	[SV_DHCPCD_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[SV_DHCPCD_ACTION] = { .name="action", .type=BLOBMSG_TYPE_STRING },
};

/* Enum for EGI policy order */
enum {
	IWLIST_SCANNING_ID,
	IWLIST_SCANNING_IFNAME,
	__IWLIST_SCANNING_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy iwlist_scanning_policy[] =
{
	[IWLIST_SCANNING_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[IWLIST_SCANNING_IFNAME] = { .name="ifname", .type=BLOBMSG_TYPE_STRING },
};

/* Enum for EGI policy order */
enum {
	START_PERIODIC_ID,
	START_PERIODIC_INTERVAL,
	__START_PERIODIC_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy start_periodic_policy[] =
{
	[START_PERIODIC_ID] = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[START_PERIODIC_INTERVAL] = { .name="interval", .type=BLOBMSG_TYPE_INT32 },
};

/* Enum for EGI policy order */
enum {
	LSLOCKS_ID,
	LSLOCKS_FILENAME,
	__LSLOCKS_MAX,
};
 
/* Ubus Policy */
const struct blobmsg_policy lslocks_policy[] =
{
	[LSLOCKS_ID]  = { .name="id", .type=BLOBMSG_TYPE_INT32 },
	[LSLOCKS_FILENAME] = { .name="filename", .type=BLOBMSG_TYPE_STRING },
};
 
/* Ubus Methods */
const struct ubus_method netaid_methods[] =
{
	UBUS_METHOD("interface_down", interface_down_handler, interface_down_policy),
	UBUS_METHOD("interface_up", interface_up_handler, interface_up_policy),
	UBUS_METHOD("ifdown", ifdown_handler, ifdown_policy),
	UBUS_METHOD("ifup", ifup_handler, ifup_policy),
	UBUS_METHOD("ipaddr_flush", ipaddr_flush_handler, ipaddr_flush_policy),
	UBUS_METHOD("disconnect", disconnect_handler, disconnect_policy),
	UBUS_METHOD("wired_connection", wired_connection_handler, wired_connection_policy),
	UBUS_METHOD("wpa_passphrase", wpa_passphrase_handler, wpa_passphrase_policy),
	UBUS_METHOD("wpa_supplicant", wpa_supplicant_handler, wpa_supplicant_policy),
	UBUS_METHOD("wireless_connection", wireless_connection_handler, wireless_connection_policy),
	UBUS_METHOD("uninstall", uninstall_handler, uninstall_policy),
	UBUS_METHOD("scan_active_wifis", scan_active_wifis_handler, scan_active_wifis_policy),
	UBUS_METHOD("nl_monitor_touch", nl_monitor_touch_handler, nl_monitor_touch_policy),
	UBUS_METHOD("iwlist_scanning", iwlist_scanning_handler, iwlist_scanning_policy),
	UBUS_METHOD("start_periodic", start_periodic_handler, start_periodic_policy),
	UBUS_METHOD("lslocks", lslocks_handler, lslocks_policy),		
};
 
/* Ubus object type */
struct ubus_object_type netaid_obj_type =
	UBUS_OBJECT_TYPE("ering.netaid", netaid_methods);
 
/* Ubus object */
struct ubus_object netaid_obj=
{
	.name = "ering.netaid", 	/* with APP name */
	.type = &netaid_obj_type,
	.methods = netaid_methods,
	.n_methods = ARRAY_SIZE(netaid_methods),
	//.path= /* useless */
};

static
void wait_on_ifupdown(int time)
{
   FILE *fp = NULL;
   char ls[128] = {0};
   
   fp = popen("ls /var/run/network | grep .pid", "r");
    if(!fp) {
       printf("Sorry, there was an error opening the pipe\n");
       exit(EXIT_FAILURE);
    }
   
    usleep(time);
    if(fgets(ls, 128, fp)!=NULL) {
	   char path[128] = {0};
	   ls[strcspn(ls, "\n")] = 0;
	   snprintf(path, sizeof(path), "/var/run/network/");
	   strcat(path, ls);
       while(!access(path, F_OK )) sleep(1);
    }
    pclose(fp);
}  

/**
 * \brief Callback function for ubus interface_down method handling
 */
int interface_down_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   sbuf_t s;
   
   /* for parsed attr */
   struct blob_attr *tb[__INTERFACE_DOWN_MAX];
    
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(interface_down_policy, ARRAY_SIZE(interface_down_policy), tb, blob_data(msg_t), blob_len(msg_t));

   const char *ifname = blobmsg_data(tb[INTERFACE_DOWN_IFNAME]);
   
   interface_down(ifname);

   sbuf_init(&s);  
   sbuf_concat(&s, 3, "ip link set ", ifname, " down");
 
   /* send a reply msg to the caller for information */
   blob_buf_init(&bb_t, 0);
   blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", s.buf);
   ubus_send_reply(ctx, req, bb_t.head);
 
   /* 	-----  reply results to the caller ----- */
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   
   sbuf_free(&s);

   return 0;
}

/**
 * \brief Callback function for ubus interface_up method handling
 */
int interface_up_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   sbuf_t s;
   
   /* for parsed attr */
   struct blob_attr *tb[__INTERFACE_UP_MAX];
    
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(interface_up_policy, ARRAY_SIZE(interface_up_policy), tb, blob_data(msg_t), blob_len(msg_t));

   const char *ifname = blobmsg_data(tb[INTERFACE_DOWN_IFNAME]);
   
   interface_up(ifname);

   sbuf_init(&s);  
   sbuf_concat(&s, 3, "ip link set ", ifname, " up");
 
   /* send a reply msg to the caller for information */
   blob_buf_init(&bb_t, 0);
   blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", s.buf);
   ubus_send_reply(ctx, req, bb_t.head);
 
   /* 	-----  reply results to the caller ----- */
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   
   sbuf_free(&s);

   return 0;
}

/**
 * \brief Callback function for ubus ifdown method handling
 */
int ifdown_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   int rc = 0;
   char msg[64] = {0};
   int tty = 0;
   char pty[16] = {0};
   char *arr = NULL;   
   sbuf_t s;
   bool redirect_ok = false;
   
   /* for parsed attr */
   struct blob_attr *tb[__IFDOWN_MAX];
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(ifdown_policy, ARRAY_SIZE(ifdown_policy), tb, blob_data(msg_t), blob_len(msg_t));

   const char *ifname = blobmsg_data(tb[IFDOWN_IFNAME]);   

   if(ifquery(ifname)) 
	  return 0;
   
   if (tb[IFDOWN_TTY] != NULL) {
	   redirect_ok = true;
       tty = blobmsg_get_u32(tb[IFDOWN_TTY]);
   }
	  
   if (!redirect_ok) {
	  strncpy_t (pty, sizeof(pty), "/dev/null", strlen("/dev/null")); 
   } else if (asprintf (&arr, "%d", (u_int32_t)tty) != -1) {
	  strncpy_t (pty, sizeof(pty), "/dev/pts/", strlen("/dev/pts/")); 
      strcat (pty, arr);
      free (arr);
   }
   
   sbuf_init(&s);   
   sbuf_concat(&s, 5, "setsid sh -c '/sbin/ifdown ", ifname, " <> ", pty, " >&0 2>&1'");
   rc = system(s.buf);
   
   sbuf_init(&s);   
   sbuf_concat(&s, 3, "setsid sh -c '/bin/echo \"           ------------            \" <> ", pty, " >&0 2>&1'");
   rc = system(s.buf);
   
   if (!redirect_ok) {
       sbuf_init(&s);  
       sbuf_concat(&s, 2, "ifdown ", ifname);
 
       /* send a reply msg to the caller for information */
       blob_buf_init(&bb_t, 0);
       blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", s.buf);
       ubus_send_reply(ctx, req, bb_t.head);
   }
   
   /* 	-----  reply results to the caller ----- */
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   
   sbuf_free(&s);
   
   return rc;
}

/**
 * \brief Callback function for ubus ifup method handling
 */
int ifup_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   int rc = 0;
   char msg[64] = {0};
   int tty = 0;
   char pty[16] = {0};
   char *arr = NULL;   
   sbuf_t s;
   bool redirect_ok = false;
   
   /* for parsed attr */
   struct blob_attr *tb[__IFUP_MAX];
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(ifup_policy, ARRAY_SIZE(ifup_policy), tb, blob_data(msg_t), blob_len(msg_t));

   const char *ifname = blobmsg_data(tb[IFUP_IFNAME]);    

   if(ifquery(ifname)) 
	  return 0;
   
   if (tb[IFUP_TTY] != NULL) {
	   redirect_ok = true;
       tty = blobmsg_get_u32(tb[IFUP_TTY]);
   }
	  
   if (!redirect_ok) {
	  strncpy_t (pty, sizeof(pty), "/dev/null", strlen("/dev/null")); 
   } else if (asprintf (&arr, "%d", (u_int32_t)tty) != -1) {
	  strncpy_t (pty, sizeof(pty), "/dev/pts/", strlen("/dev/pts/")); 
      strcat (pty, arr);
      free (arr);
   }
   
   sbuf_init(&s);   
   sbuf_concat(&s, 5, "setsid sh -c '/sbin/ifup ", ifname, " <> ", pty, " >&0 2>&1'");
   rc = system(s.buf);
   
   sbuf_init(&s);   
   sbuf_concat(&s, 3, "setsid sh -c '/bin/echo \"           ------------            \" <> ", pty, " >&0 2>&1'");
   rc = system(s.buf);
   
   if (!redirect_ok) {
       sbuf_init(&s);  
       sbuf_concat(&s, 2, "ifup ", ifname);
 
       /* send a reply msg to the caller for information */
       blob_buf_init(&bb_t, 0);
       blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", s.buf);
       ubus_send_reply(ctx, req, bb_t.head);
   }
 
   /* 	-----  reply results to the caller ----- */
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   
   sbuf_free(&s);
   
   return rc;
}

/**
 * \brief Callback function for ubus ipaddr_flush method handling
 */
int ipaddr_flush_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   sbuf_t wired, wireless;

   /* for parsed attr */
   struct blob_attr *tb[__IPADDR_FLUSH_MAX];
   
   sbuf_init(&wired);
   sbuf_init(&wireless);
   get_interfaces(&wired, &wireless);
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(ipaddr_flush_policy, ARRAY_SIZE(ipaddr_flush_policy), tb, blob_data(msg_t), blob_len(msg_t));
   
   const char *ifname = blobmsg_data(tb[IPADDR_FLUSH_IFNAME]);
   
   // For those cases where the ethernet cable was unplugged, 
   // but hadn't been disconnected by software previously
   if(!strcmp(ifname, wireless.buf)) {
      if(get_interface_status(wired.buf)) {
         interface_down(wired.buf);
         usleep(2500);
         interface_up(wired.buf);
      }
   }
   else if(!strcmp(ifname, wired.buf)) {
      if(get_interface_status(wireless.buf)) {
         interface_down(wireless.buf);
         usleep(2500);
         interface_up(wireless.buf);
      }
   }
   
   sbuf_free(&wired);
   sbuf_free(&wireless);
   
   sleep(1);

   kill_all_processes();
   ipaddr_flush(ifname);

   return 0;
}

/**
 * \brief Callback function for ubus disconnect method handling
 */
int disconnect_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   int rc __attribute__((unused)) = 0;
   char msg[64] = {0};
   int tty = 0;
   char pty[16] = {0};
   char *arr = NULL;   
   sbuf_t s;
   char sl[]="/proc/1/exe";
   char buf[32]={0};
   FILE *pfin = NULL;
   pid_t pid;
   int wstatus;
   ssize_t sz = -1;
   bool redirect_ok = false;
   
   if (lslocks (m_config->lockfile_path))
      return -1;
   else
      lock_file (m_config->lockfile_path, 1);
   
   /* for parsed attr */
   struct blob_attr *tb[__DISCONNECT_MAX];
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(disconnect_policy, ARRAY_SIZE(disconnect_policy), tb, blob_data(msg_t), blob_len(msg_t));
   
   if (tb[DISCONNECT_TTY] != NULL) {
	   redirect_ok = true;
       tty = blobmsg_get_u32(tb[DISCONNECT_TTY]);
   }
	  
   if (!redirect_ok) {
	  strncpy_t (pty, sizeof(pty), "/dev/null", strlen("/dev/null")); 
   } else if (asprintf (&arr, "%d", (u_int32_t)tty) != -1) {
	  strncpy_t (pty, sizeof(pty), "/dev/pts/", strlen("/dev/pts/")); 
      strcat (pty, arr);
      free (arr);
   }
   
   stop_periodic();
   
   sz = readlink(sl, buf, sizeof(buf));   
   if (buf && !strcmp(buf, "/sbin/runit")) {
	  sbuf_init(&s);        
      sbuf_addstr(&s, "sv stop dhcpcd");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }
   }
   
   kill_all_processes();
   
   sbuf_init(&s);
   while (*iproute() != 0) {
	   sbuf_t cmd;
	   const char *ifname = iproute();
	   if(s.len == 0)  
          sbuf_concat(&s, 2, "disconnect ", ifname);
   
       sbuf_init(&cmd);   
       sbuf_concat(&cmd, 5, "setsid sh -c '/sbin/ifdown ", ifname, " <> ", pty, " >&0 2>&1'");
       pfin = epopen(cmd.buf, &pid);
       if(pfin) { 
          fclose(pfin);
          waitpid(pid, &wstatus, 0);
       }
       
       ipaddr_flush(ifname);   
       interface_down(ifname);
       interface_up(ifname);
       sbuf_free(&cmd);
   }
   
   if (!redirect_ok) {
       /* send a reply msg to the caller for information */
       blob_buf_init(&bb_t, 0);
       blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", s.buf);
       ubus_send_reply(ctx, req, bb_t.head);
   }
    
   /* 	-----  reply results to the caller ----- */
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   
   sbuf_free(&s);
   
   lock_file (m_config->lockfile_path, 0);
   
   return rc;
}

/**
 * \brief Callback function for ubus wired_connection method handling
 */
int wired_connection_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{ 
   int rc __attribute__((unused)) = 0;
   int tty = 0;
   char pty[16];
   char msg[128]={0};
   char *arr = NULL;
   sbuf_t s;
   FILE *pfin = NULL;
   pid_t pid;
   int wstatus;
   bool redirect_ok = false;
   
   if (lslocks (m_config->lockfile_path))
      return -1;
   else
      lock_file (m_config->lockfile_path, 1);
   
   if (*iproute() != 0 ) {
	  sbuf_init(&s);   
      sbuf_concat(&s, 2, "Already connected to ", iproute());
	  goto Finish_wired_connection;
   }
   
   /* for parsed attr */
   struct blob_attr *tb[__WIRED_CONNECTION_MAX];
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(wired_connection_policy, ARRAY_SIZE(wired_connection_policy), tb, blob_data(msg_t), blob_len(msg_t)); 
  
   const char *ifname = blobmsg_data(tb[WIRED_CONNECTION_IFNAME]);
   
   if (tb[WIRED_CONNECTION_TTY] != NULL) {
	   redirect_ok = true;
       tty = blobmsg_get_u32(tb[WIRED_CONNECTION_TTY]);
   }
	  
   if (!redirect_ok) {
	  strncpy_t (pty, sizeof(pty), "/dev/null", strlen("/dev/null")); 
   } else if (asprintf (&arr, "%d", (u_int32_t)tty) != -1) {
	  strncpy_t (pty, sizeof(pty), "/dev/pts/", strlen("/dev/pts/")); 
      strcat (pty, arr);
      free (arr);
   }
   
   /* ipaddr_flush */
   
   if(get_interface_status(ifname)) {
      interface_down(ifname);
      usleep(2500);
      interface_up(ifname);
      sleep(1);
   }

   kill_all_processes();
   ipaddr_flush(ifname);
   
   /* ip link set <ifname> down */
   
   interface_down(ifname);
   
   /* ifdown <ifaname> */
   
   if (ifquery(ifname)) {
	   
      sbuf_init(&s);   
      sbuf_concat(&s, 5, "setsid sh -c '/sbin/ifdown ", ifname, " <> ", pty, " >&0 2>&1'");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }

      if (wstatus != 0)
         goto Finish_wired_connection;
   
      sbuf_init(&s);   
      sbuf_concat(&s, 3, "setsid sh -c '/bin/echo \"           ------------            \" <> ", pty, " >&0 2>&1'");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }

      if (wstatus != 0)
         goto Finish_wired_connection;
      
   }
   
   /* ip link set <ifname> up */
   
   interface_up (ifname);
   
   /* ifup */
 
   sleep(1);

   if (!ifquery(ifname)) {
   
      sbuf_init(&s);   
      sbuf_concat(&s, 5, "setsid sh -c '/sbin/ifup ", ifname, " <> ", pty, " >&0 2>&1'");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }

      if (wstatus != 0)
         goto Finish_wired_connection;
         
      rc = wstatus;   
   
      sbuf_init(&s);   
      sbuf_concat(&s, 3, "setsid sh -c '/bin/echo \"           ------------            \" <> ", pty, " >&0 2>&1'");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }
   }
   
   sbuf_init(&s);   
   sbuf_concat(&s, 2, "Connected to ", ifname);

Finish_wired_connection:

   lock_file (m_config->lockfile_path, 0);
   
   if (!redirect_ok) {
       /* send a reply msg to the caller for information */
       blob_buf_init(&bb_t, 0);
       blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", s.buf);
       ubus_send_reply(ctx, req, bb_t.head);
   }
   
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   
   sbuf_free(&s);

   return rc;
}

/**
 * \brief Callback function for ubus wpa_passphrase method handling
 */
int wpa_passphrase_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   int rc = 0;   
   sbuf_t s;
   
   /* for parsed attr */
   struct blob_attr *tb[__WPA_PASSPHRASE_MAX];
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(wpa_passphrase_policy, ARRAY_SIZE(wpa_passphrase_policy), tb, blob_data(msg_t), blob_len(msg_t));   

   const char *essid = blobmsg_data(tb[WPA_PASSPHRASE_ESSID]);
   const char *passwd = blobmsg_data(tb[WPA_PASSPHRASE_PASSWD]);
   const char *filename = blobmsg_data(tb[WPA_PASSPHRASE_FILENAME]);
   
   if(*passwd != 0) {
      
      sbuf_init(&s);      
      sbuf_concat(&s, 7, "/usr/bin/wpa_passphrase \"", essid, "\" \"", passwd, 
                         "\"  | sudo tee -a \"", filename, "\" > /dev/null");      
      rc = system(s.buf);
      
   } else {
   
      FILE *fp;   
      fp = fopen(filename, "w");
      if(!fp) {
         printf("Cannot open the \"%s\" config file.\n", filename);
         return 1;
      }
      
      fprintf( fp, "network={\n\tssid=\"%s\"\n\tkey_mgmt=NONE\n}", essid);
      fclose(fp);
   }
   
   if(access(filename, F_OK) == 0) {
	  const char *dname = dirname (strdup(filename));
	  if (!strcmp(dname, WIFI_DIR)) { 
         sbuf_init(&s);
         sbuf_concat(&s, 3, "chmod -R 644 \"", filename, "\" > /dev/null");  
         rc = system(s.buf);
      }   
   }
   
   sbuf_free(&s);
    
   /* send a reply msg to the caller for information */
   blob_buf_init(&bb_t, 0);
   blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", "wpa_passphrase");
   ubus_send_reply(ctx, req, bb_t.head);
 
   /* 	-----  reply results to the caller ----- */
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   
   return rc;
}

/**
 * \brief Callback function for ubus wpa_supplicant method handling
 */
int wpa_supplicant_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   int rc = 0;
   int tty = 0;
   char pty[16] = {0};
   char *arr = NULL;      
   char path[1024]={0};
   sbuf_t s;
   FILE *pfin = NULL;
   pid_t pid;
   int wstatus;
   bool redirect_ok = false;
   
   /* for parsed attr */
   struct blob_attr *tb[__WPA_SUPPLICANT_MAX];
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(wpa_supplicant_policy, ARRAY_SIZE(wpa_supplicant_policy), tb, blob_data(msg_t), blob_len(msg_t));   
   
   const char *ifname = blobmsg_data(tb[WPA_SUPPLICANT_IFNAME]);
   const char *filename = blobmsg_data(tb[WPA_SUPPLICANT_FILENAME]);
   
   if (tb[WPA_SUPPLICANT_TTY] != NULL) {
	   redirect_ok = true;
       tty = blobmsg_get_u32(tb[WPA_SUPPLICANT_TTY]);
   }
	  
   if (!redirect_ok) {
	  strncpy_t (pty, sizeof(pty), "/dev/null", strlen("/dev/null")); 
   } else if (asprintf (&arr, "%d", (u_int32_t)tty) != -1) {
	  strncpy_t (pty, sizeof(pty), "/dev/pts/", strlen("/dev/pts/")); 
      strcat (pty, arr);
      free (arr);
   }
   
   sbuf_init(&s);
   sbuf_concat(&s, 7, "setsid sh -c '/sbin/wpa_supplicant -B -i\"", ifname, 
                        "\" -c\"", filename, "\" <> ", pty, " >&0 2>&1'");
   
   pfin = epopen(s.buf, &pid);
   if(pfin) { 
      fclose(pfin);
      waitpid(pid, &wstatus, 0);
   }
   sbuf_free(&s);
   
   if (!filename)
      unlink(path);
 
   if (!redirect_ok) {
       /* send a reply msg to the caller for information */
       blob_buf_init(&bb_t, 0);
       blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", "wpa_supplicant");
       ubus_send_reply(ctx, req, bb_t.head);
   } 
 
   /* 	-----  reply results to the caller ----- */
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   
   return rc;
}

/**
 * \brief Callback function for ubus wireless_connection method handling
 */
int wireless_connection_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{ 
   int rc __attribute__((unused)) = 0;
   int tty = 0;
   char pty[16];      
   char path[1024];
   char msg[128]={0};
   sbuf_t s;
   char *arr = NULL;
   FILE *pfin = NULL;
   pid_t pid;
   int wstatus;
   bool redirect_ok = false;
   bool pass_filter_ok = false;
   bool give_permissions_ok = false;
   
   if (lslocks (m_config->lockfile_path))
      return -1;
   else
      lock_file (m_config->lockfile_path, 1);
   
   if (*iproute() != 0 ) {
	  sbuf_init(&s);   
      sbuf_concat(&s, 2, "Already connected to ", iproute());
	  goto Finish_wireless_connection;
   }
   
   /* for parsed attr */
   struct blob_attr *tb[__WIRELESS_CONNECTION_MAX];
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(wireless_connection_policy, ARRAY_SIZE(wireless_connection_policy), tb, blob_data(msg_t), blob_len(msg_t)); 
  
   const char *ifname = blobmsg_data(tb[WIRELESS_CONNECTION_IFNAME]);
   const char *essid = blobmsg_data(tb[WIRELESS_CONNECTION_ESSID]);
   const char *passwd = blobmsg_data(tb[WIRELESS_CONNECTION_PASSWD]);
   const char *filename = blobmsg_data(tb[WIRELESS_CONNECTION_FILENAME]);
   
   if (tb[WIRELESS_CONNECTION_TTY] != NULL) {
	   redirect_ok = true;
       tty = blobmsg_get_u32(tb[WIRELESS_CONNECTION_TTY]);
   }
	  
   if (!redirect_ok) {
	  strncpy_t (pty, sizeof(pty), "/dev/null", strlen("/dev/null")); 
   } else if (asprintf (&arr, "%d", (u_int32_t)tty) != -1) {
	  strncpy_t (pty, sizeof(pty), "/dev/pts/", strlen("/dev/pts/")); 
      strcat (pty, arr);
      free (arr);
   }
   
   /* passwd may be empty, bear in mind open wifis */
   if (*filename != 0) {
	  strncpy_t(path, sizeof(path), filename, strlen(filename));
   }
   else {
	  sbuf_t random;
	  sbuf_init(&random); 
      strncpy_t(path, sizeof(path), "/tmp/simple-netaid_", strlen("/tmp/simple-netaid_"));
      sbuf_random(&random, 5);
      strcat(path, random.buf);
      sbuf_free(&random);
   }
   
   const char *dname = dirname (strdup(path));
   if (!strcmp(dname, WIFI_DIR) || !strncmp(path, PREFIX, strlen(PREFIX)))
      pass_filter_ok = true;
   
   /* ipaddr_flush */
   if(get_interface_status(ifname)) {
      interface_down(ifname);
      usleep(2500);
      interface_up(ifname);
      sleep(1);
   }

   kill_all_processes();
   ipaddr_flush(ifname);
   
   /* ip link set <ifname> down */
   
   interface_down(ifname);
   
   /* ifdown <ifname> */
   if (ifquery(ifname)) {
	   
      sbuf_init(&s);   
      sbuf_concat(&s, 5, "setsid sh -c '/sbin/ifdown ", ifname, " <> ", pty, " >&0 2>&1'");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }
   }
   
   /* ip link set <ifname> up */
   interface_up (ifname);
   
   /* wpa_passphrase */
   if (pass_filter_ok && *passwd != 0 && access(path, F_OK) != 0) {
      sbuf_init(&s);
      sbuf_concat(&s, 7, "/usr/bin/wpa_passphrase \"", essid, "\" \"", passwd, 
                         "\"  | sudo tee -a \"", path, "\" > /dev/null");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }

      if (wstatus != 0)
         goto Finish_wireless_connection;
	  
	  pass_filter_ok = true;
      
      /* Give right permissions to the already installed file */
      sbuf_init(&s);
      sbuf_concat(&s, 3, "chmod -R 644 \"", path, "\" > /dev/null");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }

      if (wstatus != 0)
         goto Finish_wireless_connection;
   }
   else if (*passwd == 0 && *filename == 0) {
   
      FILE *fp = fopen(path, "w");
      if(!fp) {
         printf("Cannot open the \"%s\" config file.\n", path);
         return 1;
      }
      
      fprintf( fp, "network={\n\tssid=\"%s\"\n\tkey_mgmt=NONE\n}", essid);
      fclose(fp);
   }

   /* wpa_supplicant */
   sbuf_init(&s);
   sbuf_concat(&s, 7, "setsid sh -c '/sbin/wpa_supplicant -B -i\"", ifname, 
                        "\" -c\"", path, "\" <> ", pty, " >&0 2>&1'");
   pfin = epopen(s.buf, &pid);
   if(pfin) { 
      fclose(pfin);
      waitpid(pid, &wstatus, 0);
   }

   if (wstatus != 0)
      goto Finish_wireless_connection;      
   
   // if path is in /tmp
   if (*filename == 0 && !strncmp(path, PREFIX, strlen(PREFIX)))
      unlink(path);
   
   /* ifup */
   
   sleep(1);

   if (!ifquery(ifname)) {
   
      sbuf_init(&s);   
      sbuf_concat(&s, 5, "setsid sh -c '/sbin/ifup ", ifname, " <> ", pty, " >&0 2>&1'");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }

      if (wstatus != 0)
         goto Finish_wireless_connection;
      
      rc = wstatus;
   
      sbuf_init(&s);   
      sbuf_concat(&s, 3, "setsid sh -c '/bin/echo \"           ------------            \" <> ", pty, " >&0 2>&1'");
      pfin = epopen(s.buf, &pid);
      if(pfin) { 
         fclose(pfin);
         waitpid(pid, &wstatus, 0);
      }     
   }
   
   sbuf_init(&s);   
   sbuf_concat(&s, 2, "Connected to ", ifname);
   
   /* Remove installed file in case of failed connection attempts */
   if(pass_filter_ok && *iproute() == 0)
      unlink (path); 

Finish_wireless_connection:

   lock_file (m_config->lockfile_path, 0);
   
   if (!redirect_ok) {
       /* send a reply msg to the caller for information */
       blob_buf_init(&bb_t, 0);
       blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", s.buf);
       ubus_send_reply(ctx, req, bb_t.head);
   }
   
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   
   sbuf_free(&s);

   return rc;
}

/**
 * \brief Callback function for ubus uninstall method handling
 */
int uninstall_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   char msg[64] = {0};
   char *aux = NULL;

   /* for parsed attr */
   struct blob_attr *tb[__UNINSTALL_MAX]; 
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(uninstall_policy, ARRAY_SIZE(uninstall_policy), tb, blob_data(msg_t), blob_len(msg_t));

   const char *filename = blobmsg_data(tb[UNINSTALL_FILENAME]);
   aux = dirname(strdup(filename));

   if(!strcmp(aux, WIFI_DIR)) {         
      unlink(filename);
      if(access(filename, F_OK) != -1) {
         log_error("%s: Error deleting %s file.\n", progname, filename);
      }
   }
    
   free(aux);

   return 0;
}

/**
 * \brief Callback function for ubus scan_active_wifis method handling
 */
int scan_active_wifis_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   int rc = 0, format = 0;
   
   if (lslocks (m_config->active_wifis_lockfile_path))
      return -1;
   else
      lock_file (m_config->active_wifis_lockfile_path, 1);

   /* for parsed attr */
   struct blob_attr *tb[__SCAN_ACTIVE_WIFIS_MAX]; 
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(scan_active_wifis_policy, ARRAY_SIZE(scan_active_wifis_policy), tb, blob_data(msg_t), blob_len(msg_t));
   
   if (tb[SCAN_ACTIVE_WIFIS_FORMAT] != NULL) {
       format = blobmsg_get_u32(tb[SCAN_ACTIVE_WIFIS_FORMAT]);
   }
   
   if (format == 0) {
       sbuf_t s; // Don't use __cleanbuf__ attribute together with ubus_send_replay()
       sbuf_init(&s);
       rc = scan_active_wifis(blobmsg_data(tb[SCAN_ACTIVE_WIFIS_IFNAME]), &s);  /* send a reply msg to the caller for information */
       blob_buf_init(&bb_t, 0);
       blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", s.buf);
       ubus_send_reply(ctx, req, bb_t.head);
   
       ubus_defer_request(ctx, req, &req_data);
       ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);

       sbuf_free(&s);
   }
   else {
 
	   blob_buf_init(&bb_t, 0);
	   int count = iwlist (blobmsg_data(tb[SCAN_ACTIVE_WIFIS_IFNAME]), &bb_t);
       ubus_send_reply(ctx, req, bb_t.head);
   
       ubus_defer_request(ctx, req, &req_data);
       ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);
   }
   
   lock_file (m_config->active_wifis_lockfile_path, 0);

   return rc;
}

/**
 * \brief Callback function for ubus nl_monitor_touch method handling
 */
int nl_monitor_touch_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   int rc = 0, uid, gid;
   char *uid_str = NULL;
   char *gid_str = NULL;
   int fd = -1;   /* file descriptor */
   const char *filename;
   sbuf_t s;

   /* for parsed attr */
   struct blob_attr *tb[__NL_MONITOR_TOUCH_MAX]; 
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(nl_monitor_touch_policy, ARRAY_SIZE(nl_monitor_touch_policy), tb, blob_data(msg_t), blob_len(msg_t));

   uid = blobmsg_get_u32(tb[NL_MONITOR_TOUCH_UID]);
   gid = blobmsg_get_u32(tb[NL_MONITOR_TOUCH_GID]);
   filename = blobmsg_data(tb[NL_MONITOR_TOUCH_FILENAME]);
    
   if(asprintf(&uid_str, "%d", uid) == -1) {
      printf("Unable to get uid: %s\n", strerror(errno));
      exit(EXIT_FAILURE);
   }
    
   if(asprintf(&gid_str, "%d", gid) == -1) {
      printf("Unable to get uid: %s\n", strerror(errno));
      free(uid_str);
      exit(EXIT_FAILURE);
   }
   
   sbuf_init(&s);
   sbuf_concat(&s, 4, "/tmp/snetaid_", uid_str, "_", gid_str);
   rc = mkdir(s.buf, 0755);
   if (rc == -1) {
      switch (errno) {
          case EACCES :
              log_error("%s: parent directory does not allow write\n", progname);free(uid_str);
	          goto Free;
              return 1;
          case EEXIST:
              log_info("%s: pathname '%s' already exists\n", progname, s.buf);
              break;
          case ENAMETOOLONG:
              log_error("%s: pathname '%s' is too long\n", progname, s.buf);
	          goto Free;
              return 1;
          default:
              log_error("%s: mkdir() '%s'", progname, s.buf);
	          goto Free;
              return 1;
      }
   }
    
   chown(s.buf, uid, gid);
   
   // touch the file      
   sbuf_concat(&s, 2, "/", filename);
   fd = open(s.buf, O_CREAT | S_IRUSR | S_IWUSR, O_RDWR);
   if (fd == -1) {
	  sbuf_free(&s); 
      log_error("%s: Unable to touch file %s", progname, s.buf);
      return 1;
   }
      
   fchmod(fd, 0755);
   fchown(fd, uid, gid);
   close(fd);
   
Free:
   sbuf_free(&s);
   free(uid_str);
   free(gid_str);
   return 0;
}

/**
 * \brief Callback function for ubus sv_dhcpcd method handling
 */
 /*
int sv_dhcpcd_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   FILE *pfin = NULL;
   pid_t pid;
   int wstatus;
   sbuf_t s;

   // for parsed attr
   struct blob_attr *tb[__SV_DHCPCD_MAX]; 
 
   // Parse blob_msg from the caller to request policy
   blobmsg_parse(sv_dhcpcd_policy, ARRAY_SIZE(sv_dhcpcd_policy), tb, blob_data(msg_t), blob_len(msg_t));

   const char *action = blobmsg_data(tb[SV_DHCPCD_ACTION]);
   
   if(strcmp(action, "start") && strcmp(action, "stop"))
      return 1;
   
   sbuf_init(&s);        
   sbuf_concat(&s, 3, "sv ", action, " dhcpcd");
   pfin = epopen(s.buf, &pid);
   if(pfin) { 
      fclose(pfin);
      waitpid(pid, &wstatus, 0);
   }
   
   sbuf_free(&s);

   return wstatus;
}
*/

/**
 * \brief Callback function for ubus iwlist_scanning method handling
 */
int iwlist_scanning_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   int rc = 0, i = 0, j = 0, k = 0;
   sbuf_t s; // Don't use __cleanbuf__ attribute together with ubus_send_replay()
   sbuf_t aux;
   sbuf_t address;
   sbuf_t essid;
   sbuf_t quality;
   const char *ifname;
 
   blob_buf_init(&bb_t, 0);
 
   struct blob_attr *tb[__IWLIST_SCANNING_MAX]; 
   blobmsg_parse(iwlist_scanning_policy, ARRAY_SIZE(iwlist_scanning_policy), tb, blob_data(msg_t), blob_len(msg_t));
   ifname = blobmsg_data(tb[IWLIST_SCANNING_IFNAME]);
   
   sbuf_init(&s);

   rc = scan_active_wifis(ifname, &s);
   
   while(s.buf[i] != '\0') {  
        sbuf_init(&aux);
        while (s.buf[i] != '|' && s.buf[i] != '\0') {
			sbuf_addch(&aux, s.buf[i]);
			i++;
		}
   const char *ifname;
 
   blob_buf_init(&bb_t, 0);
 
   struct blob_attr *tb[__IWLIST_SCANNING_MAX]; 
   blobmsg_parse(iwlist_scanning_policy, ARRAY_SIZE(iwlist_scanning_policy), tb, blob_data(msg_t), blob_len(msg_t));
   ifname = blobmsg_data(tb[IWLIST_SCANNING_IFNAME]);
		sbuf_addch(&aux, '\0');
		if(j==0) {
            sbuf_init(&address);
            sbuf_addstr(&address, aux.buf);
		    j++;
		}
		else if(j==1) {
            sbuf_init(&essid);
            sbuf_addstr(&essid, aux.buf);
			j++;
		}
		else if(j==2) {
            sbuf_init(&quality);
            sbuf_addstr(&quality, aux.buf);
		    j=0;
		    k++;
		    void *tbl;
		    char cell[10];
            rc = sprintf(cell, "Cell %d", k);
	        tbl = blobmsg_open_table(&bb_t, cell);
	        blobmsg_add_string(&bb_t, "Address", address.buf);
	        blobmsg_add_string(&bb_t, "Essid", essid.buf);
	        blobmsg_add_string(&bb_t, "Quality", quality.buf);
	        blobmsg_close_table(&bb_t, tbl);
		}
		i++;
   }
	
   //blobmsg_format_json(bb_t.head, true);
   ubus_send_reply(ctx, req, bb_t.head);
   
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);

   sbuf_free(&s);
   sbuf_free(&aux);
   sbuf_free(&address);
   sbuf_free(&essid);
   sbuf_free(&quality);
   
   return rc;
}

/**
 * \brief Callback function for ubus start_periodic method handling
 */
int start_periodic_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   int interval;

   /* for parsed attr */
   struct blob_attr *tb[__START_PERIODIC_MAX]; 
 
   /* Parse blob_msg from the caller to request policy */
   blobmsg_parse(start_periodic_policy, ARRAY_SIZE(start_periodic_policy), tb, blob_data(msg_t), blob_len(msg_t));
   
   interval = blobmsg_get_u32(tb[START_PERIODIC_INTERVAL]);
   
   if(interval>0)
      start_periodic(interval);
   else
      stop_periodic();

   return 0;
}

/**
 * \brief Callback function for ubus lslocks method handling
 */
int lslocks_handler( struct ubus_context *ctx, struct ubus_object *obj,
			  struct ubus_request_data *req, const char *method,
			  struct blob_attr *msg_t )
{
   sbuf_t s; // Don't use __cleanbuf__ attribute together with ubus_send_replay()
   FILE *pf = NULL;
   const char *filename;
 
   blob_buf_init(&bb_t, 0);
 
   struct blob_attr *tb[__IWLIST_SCANNING_MAX]; 
   blobmsg_parse(lslocks_policy, ARRAY_SIZE(lslocks_policy), tb, blob_data(msg_t), blob_len(msg_t));
   filename = blobmsg_data(tb[LSLOCKS_FILENAME]);
   
   sbuf_init(&s);
   
   if (!strcmp(filename, "lockfile"))
      sbuf_concat(&s, 3, "/usr/bin/lslocks | /bin/grep ", m_config->lockfile_path, "$"); 
   else if (!strcmp(filename, "active_wifis_lockfile"))
      sbuf_concat(&s, 3, "/usr/bin/lslocks | /bin/grep ", m_config->active_wifis_lockfile_path, "$");
   else {
	  sbuf_free(&s);     
      return -1;
   }
   
   pf = popen(s.buf, "r");
   if(pf) {
	   char output[1024] = {0};
       sbuf_init(&s);
	   if (fgets(output, 1024, pf) != NULL) {
           output[strcspn(output, "\n")] = 0;
	       sbuf_addstr(&s, rindex(output, ' ') + 1);
	       pclose(pf);
       }
   }
   else {
	   /* send a reply msg to the caller for information */
       blob_buf_init(&bb_t, 0);
       blobmsg_add_string(&bb_t,"Server reply - ERROR: ", "lslocks command failed");
       ubus_send_reply(ctx, req, bb_t.head);
	   return -1;
   }
 
   /* send a reply msg to the caller for information */
   blob_buf_init(&bb_t, 0);
   blobmsg_add_string(&bb_t,"Server reply - Request has been proceeded: ", s.buf);
   ubus_send_reply(ctx, req, bb_t.head);
   
   ubus_defer_request(ctx, req, &req_data);
   ubus_complete_deferred_request(ctx, req, UBUS_STATUS_OK);

   sbuf_free(&s);

   return 0;
}
